#!/usr/bin/awk -f

# Written by Gerd Stolpmann.
#----------------------------------------------------------------------
# 
# This script analyzes the dependencies between the installed packages
# and packages that can be built, and outputs the result to stdout.

# Input variables: (pass them with "awk -v")
#
# - pkglist_to_update: The list of packages to update
# - force: whether to update the packages even if there is no better
#   available version (the string "1" means "yes")
# - localbase: The path $LOCALBASE
#
# Output is printed on stdout:
#
# - 1st line: The list of packages that will be deleted (in reverse
#   topological order)
# - 2nd line: The list of packages that will be (re)built (in topological
#   order)

#----------------------------------------------------------------------

# Variables:
#
# pkg_status[name]: A string identifying the current status of the
#   package. Valid strings are:
#   "R": The package is to be replaced (delete + build)
#   "D": The package is to be deleted  (NOT USED!)
#   "B": The package is to be built
#
# pkg_avail_version[name]: The available version of the package
#
# pkg_inst_version[name]: The installed version of the package, or ""
#    if not installed
#
# pkg_depends[name]: The packages name depends on (incl. build dependencies),
#    a space-separated string

# ----------------------------------------------------------------------
#
# Algorithm:
#
# (1) Update analysis (no "force"): 
#     Iterate through pkglist_to_update, check which
#     packages need to be built or replaced, and put "B" or "R" tokens
#     into pkg_status.
#
# (2) Forced update: Iterate through pkglist_to_update, and put "B" or "R" into
#     pkg_status
#
# (3) For all packages having an "R" do recursively: Find the dependent
#     packages that are installed, and records an "R" for them, too.
#
# (4) For all packages having a "B" or "R" do recursively: Find the
#     required packages (incl. build dependencies), check whether they
#     exist, and record a "B" for them if not.
#
# (5) Extract all "R" packages, and sort them in reverse topological
#     order, and output the result as first line
#
# (6) Extract all "R" and "B" packages, and sort them in topological
#     order, and output the result as second line
#
#
# Note: Most of the checks for empty strings are necessary because
# of buggy awk implementions.

# ----------------------------------------------------------------------

function del_array(a) {
    # Deletes the array a in a portable (and slow) way
    for (k in a) {
	delete a[k];
    }
}


function quote(qs) {
    # Shell-quotes qs
    #print "qs=" qs;
    gsub(/[']/, "'\\''", qs);
    qs = "'" qs "'";
    #print "qs'= " qs;
    return qs;
}


function errmsg(msg) {
    msg = "godi_update.awk: " msg;
    system(sprintf("echo %s >&2", quote(msg)));
}


function find_pkg_dir(name) {
    # Find the package directory containing [name]. This function returns
    # only the name of the directory, without path.

    # Impl.: By convention, the package directory is the first word
    # of [name]

    split(name, words, "-");
    return words[1];
}


function base_name (name) {
    # Return the package name without version string
    sub(/-[0-9][^-]*$/, "", name);
    return name;
}


function version_suffix (name) {
    # Return the version of a package with versioned name
    p = match(name, /-[0-9][^-]*$/);
    return substr(name, p+1);
}


function avail_version(name) {
    # Returns the available version of a package
    # [name] must not contain a version number
    if (name in pkg_avail_version) {
	return pkg_avail_version[name];
    }
    else {
	if (name == "") return "";
	dir = find_pkg_dir(name);
	path = sprintf("%s/build/%s/%s", localbase, dir, name);
	pkg_exists = system(sprintf("test -d %s", quote(path)));
	if (pkg_exists != 0) {
	    errmsg("[Warning] No build instructions found for package: " name);
	    pkg_avail_version[name] = "";
	    return "";
	};
	cmd = sprintf("cd %s && %s/bin/godi_make show-var VARNAME=PKGNAME",
		      quote(path),localbase);
	success = (cmd | getline v);
	close(cmd);
	if (success > 0) {
	    v = version_suffix(v);
	} else {
	    v = "";
	};
	pkg_avail_version[name] = v;
	#printf "avail_version(%s)=%s\n", name, v;
	return v;
    }
}


function inst_version (name) {
    # Returns the installed version of a package
    # [name] must not contain a version number
    if (name in pkg_inst_version) {
	return pkg_inst_version[name];
    }
    else {
	if (name == "") return "";
	cmd = sprintf("%s/sbin/godi_info -e %s",localbase,quote(name));
	success = (cmd | getline v);
	close(cmd);
	# v is the empty string if not existing
	if (success > 0) {
	    v = version_suffix(v);
	} else {
	    v = "";
	};
	pkg_inst_version[name] = v;
	#printf "inst_version(%s)=%s\n", name, v;
	return v;
    }
}


function better_than(v1, v2) {
    # Returns true if version v1 is better (newer) than v2
    cmd = sprintf("%s/sbin/godi_admin pmatch %s %s",localbase,quote("x>" v2),quote("x-" v1));
    if (system(cmd) == 0) {
        return 1;
    };
    cmd = sprintf("%s/sbin/godi_admin pmatch %s %s",localbase,quote("x==" v2),quote("x-" v1));
    if (system(cmd) == 0) {
        # Compare package revision numbers
        rev1p = match(v1, /godi.*$/);
        rev1  = substr(v1, rev1p+4);
        rev2p = match(v2, /godi.*$/);
        rev2  = substr(v2, rev2p+4);
        return (rev1p > rev2p);
    }
    return 0;
}


function check_update() {
    # Step (1) of algorithm.
    split(pkglist_to_update, pkgs, /[ \t\r\n]+/);
    for (p in pkgs) {
	pkg=pkgs[p];
	pkg=base_name(pkg);
	# Check the installed version. If not installed, record a "B",
	# and continue.
	pkg_iv = inst_version(pkg);
	if (pkg_iv == "") {
	    pkg_status[pkg] = "B";
	    continue;
	};
	# Check the available version. If not better,
        # ignore the package
	pkg_av = avail_version(pkg);
	# Check if pkg_av is better than pkg_iv:
	if (better_than(pkg_av, pkg_iv)) {
	    # It is ok if pkg_av == ""; this is checked later again
	    pkg_status[pkg] = "R";
	}
	# else: do nothing (ignore package)
    }
}


function forced_update() {
    # Step (2) of the algorithm.
    split(pkglist_to_update, pkgs, /[ \t\r\n]+/);
    for (p in pkgs) {
	pkg=pkgs[p];
	pkg=base_name(pkg);
	if (inst_version(pkg) == "") {
	    pkg_status[pkg] = "B";
	} else {
	    pkg_status[pkg] = "R";
	}
    }
}


function replace_inst_descendents() {
    # Step (3) of the algorithm.
    for (pkg in pkg_status) {
	if (pkg_status[pkg] == "R") {
	    cmd = sprintf("%s/sbin/godi_info -q -R %s",localbase,quote(pkg));
	    while(cmd | getline dpkg) {
		if (dpkg != "") {
		    dpkg = base_name(dpkg);
		    new[dpkg] = "R";
		}
	    }
	    close(cmd);
	}
    }
    for (pkg in new) {
	pkg_status[pkg] = new[pkg];
    }
    del_array(new);
}


function check_availability() {
    # Check that all "R" packages are still available:
    for (pkg in pkg_status) {
	if (pkg_status[pkg] == "R") {
	    if (avail_version(pkg) == "") {
		# This is fatal, because "R" means to delete a package
		# that cannot be rebuilt later
		errmsg("[Error] This installed package cannot be rebuilt: " pkg);
		errmsg("However, it must be rebuilt because of dependencies.");
		exit(1);
	    };
	}
    }
}


function get_ancestors(name) {
    if (name in pkg_depends) {
	return pkg_depends[name];
    }
    if (name == "") {
        return "";
    }
    dir = find_pkg_dir(name);
    path = sprintf("%s/build/%s/%s", localbase, dir, name);
    pkg_exists = system(sprintf("test -d %s", quote(path)));
    if (pkg_exists != 0) {
	errmsg("[Warning] No build instructions found for package: " name);
	pkg_depends[name] = "";
	return "";
    };
    cmd = sprintf("cd %s && %s/bin/godi_make show-vars VARNAMES='DEPENDS BUILD_DEPENDS'",
		  quote(path),localbase);
    # How to check the exit code here?
    success1 = (cmd | getline depends);
    success2 = (cmd | getline blddeps);
    close(cmd);
    if (success1 <= 0 || success2 <= 0) {
	pkg_depends[name] = "";
	return "";
    };
    split(depends, d, /[ \t\r\n]+/);
    split(blddeps, b, /[ \t\r\n]+/);
    for (k in d) {
	ds=d[k];
	if (ds == "") continue;
	sub(/[^:]*:/, "", ds);   # remove package condition, path remains
	j=match(ds, /[^\/]+$/);
	ds=substr(ds, j);
	r[ds]=1;
    }
    for (k in b) {
	bs=b[k];
	if (bs == "") continue;
	sub(/[^:]*:/, "", bs);   # remove package condition, path remains
	j=match(bs, /[^\/]+$/);
	bs=substr(bs, j);
	r[bs]=1;
    }
    s="";
    for (pkg in r) {
	s = s " " pkg;
    }
    s = substr(s, 2);
    del_array(r);
    pkg_depends[name] = s;
    #printf "get_ancestors(%s)=%s\n", name, s;
    return s;
}


function add_required_ancestors() {
    # Step (4) of the algorithm
    for (pkg in new) {
	errmsg("[FATAL] Assertion failed!");
	exit(1);
    };

    n=1
    while (n) {
	n=0
	for (pkg1 in pkg_status) {
	    ancs_str = get_ancestors(pkg1);
	    split(ancs_str, ancs, /[ \t\r\n]+/);
	    for (k in ancs) {
		anc = ancs[k];
		if (anc == "") continue;
		#errmsg("Checking " anc);
		if (pkg_status[anc] != "B" && pkg_status[anc] != "R") {
		    # Nothing known about package
		    if (inst_version(anc) == "") {
			# TODO: Check whether version is ok
			new[anc] = "B";
			n=1;
			#errmsg("adding " anc);
		    };
		}
	    }
	}
	for (pkg in new) {
	    pkg_status[pkg] = new[pkg];
	}
	del_array(new);
    }
}


function output_topo(topo_array) {
    # Maybe we can use the tsort command.
    topo_res=""
    for (topo_pkg in topo_array) {
	# Unfortunately, we have no recursion in awk. So this strange
        # spanning tree algorithm.
	plan[1] = topo_pkg;
	planmax = 1;
	if (!visited[topo_pkg]) {
	    while (planmax >= 1) {
                # Take next of plan:
		nextpkg = plan[planmax];
		#print "nextpkg=" nextpkg;
                # Get ancestors:
		ancs_str = get_ancestors(nextpkg);
		#print "ancestors=" ancs_str;
		split(ancs_str, ancs, /[ \t\r\n]+/);
                # Add ancestors to plan:
		shrink=1;
		for (k in ancs) {
		    anc=ancs[k];
		    if (anc == "") continue;
		    if (!visited[anc] && anc in topo_array) {
			shrink=0;
			planmax++;
			plan[planmax] = anc;
		    }
		}
		if (shrink) {
		    # Node has no or only visited ancestors:
		    if (!visited[nextpkg]) {
			#print "output " nextpkg;
			topo_res = topo_res " " nextpkg;
			visited[nextpkg] = 1;
		    }
		    delete plan[planmax];
		    planmax--;
		}
	    }
	}
	    
    };
    del_array(visited);
    return substr(topo_res, 2);
}


BEGIN {
    if (force == "1") {
	forced_update();
    } else {
	check_update();
    }
    replace_inst_descendents();
    #errmsg("After replace_inst_descendents:");
    #for (pkg in pkg_status) {
    #	errmsg("- " pkg);
    #};
    check_availability();
    add_required_ancestors();
    #errmsg("After add_required_ancestors:");
    #for (pkg in pkg_status) {
    #	errmsg("- " pkg "=" pkg_status[pkg]);
    #};
    # Output line 1:
    for (pkg in pkg_status) {
	if (pkg_status[pkg] == "R") {
	    pkg_r[pkg] = 1;
	}
    }
    n = split(output_topo(pkg_r), pkg_r_rev, /[ \t\r\n]+/);
    s = ""
    for (k=n; k>=1; k--) {
	s = s " " pkg_r_rev[k];
    }
    print substr(s, 2);
    # Output line 2:
    for (pkg in pkg_status) {
	if (pkg_status[pkg] == "R" || pkg_status[pkg] == "B") {
	    pkg_rb[pkg] = 1;
	}
    };
    print output_topo(pkg_rb);
}

